import { EntityNodeAttributes, LogicNodeAttributes, QueryGraphNodes, RelationNodeAttributes } from '../model/graphology/model';
import { QueryMultiGraph } from '../model/graphology/utils';
import { BackendQueryFormat, MachineLearning, NodeStruct, QueryStruct, RelationStruct } from '../model/BackendQueryFormat';
import { QueryElementTypes, toHandleData } from '../model';
import { SerializedNode } from 'graphology-types';
import { hasCycle } from 'graphology-dag';
import Graph, { MultiGraph } from 'graphology';
import { allSimplePaths } from 'graphology-simple-path';
import { AllLogicStatement, ReferenceStatement } from '../model/logic/general';
import { ML, MLTypes, mlDefaultState } from '../../data-access/store/mlSlice';
import { QueryBuilderSettings } from '../../data-access/store/querybuilderSlice';

// export type QueryI {

// }

// Chunk extraction: traverse graph to find all paths of logic between relations and entities
const traverseEntityRelationPaths = (
  node: SerializedNode<QueryGraphNodes>,
  paths: QueryGraphNodes[][],
  currentIdx: number,
  graph: QueryMultiGraph,
  entities: SerializedNode<EntityNodeAttributes>[],
  relations: SerializedNode<RelationNodeAttributes>[],
  settings: QueryBuilderSettings
): number => {
  if (!node?.attributes) throw Error('Malformed Graph! Node has no attributes');
  // console.log(paths);

  if (!paths?.[currentIdx]) {
    // console.log('new path');
    paths.push([]);
    if (node.attributes.type === QueryElementTypes.Relation) {
      paths[currentIdx].push({ type: QueryElementTypes.Entity, x: node.attributes.x, y: node.attributes.y });
    }
  } else if (paths[currentIdx].length > 0) {
    const lastNode = paths[currentIdx][paths[currentIdx].length - 1];
    if (lastNode.type === node.attributes.type) {
      if (lastNode.type === QueryElementTypes.Entity) {
        paths[currentIdx].push({
          type: QueryElementTypes.Relation,
          x: node.attributes.x,
          y: node.attributes.x,
          depth: { min: settings.depth.min, max: settings.depth.max },
          direction: 'both',
        });
      } else {
        paths[currentIdx].push({ type: QueryElementTypes.Entity, x: node.attributes.x, y: node.attributes.x });
      }
    }
  }
  paths[currentIdx].push(node.attributes);

  const edges = graph.edges.filter(
    (n) =>
      n?.attributes?.sourceHandleData.nodeType !== QueryElementTypes.Logic &&
      n?.attributes?.targetHandleData.nodeType !== QueryElementTypes.Logic
  );
  let connections = edges.filter((e) => e.source === node.key);
  if (connections.length === 0) {
    if (node.attributes.type === QueryElementTypes.Relation) {
      paths[currentIdx].push({ type: QueryElementTypes.Entity, x: node.attributes.x, y: node.attributes.x });
    }
    return 0;
  }
  // console.log('connections', connections);

  const nodesToRight = connections
    .map((c, i) => {
      const rightNodeHandleData = c.attributes?.targetHandleData;
      if (!rightNodeHandleData) throw Error('Malformed Graph! One or more edges of a relation node do not exist');

      // console.log('nodesToRight', c);
      // console.log('entities', entities);
      if (paths[currentIdx].length > 10 || currentIdx > 10)
        throw Error('Malformed Graph! One or more edges of a relation node do not exist');
      const rightNode =
        rightNodeHandleData.nodeType === QueryElementTypes.Entity
          ? entities.find((r) => r.key === c.target)
          : relations.find((r) => r.key === c.target);
      return rightNode;
    })
    .filter((n) => n !== undefined) as SerializedNode<QueryGraphNodes>[];

  let chunkOffset = 0;
  let pathBeforeTraversing = [...paths[currentIdx]];
  nodesToRight.forEach((rightNode, i) => {
    if (i > 0) {
      paths.push([...pathBeforeTraversing]); // clone previous path in case of branching
    }
    chunkOffset += traverseEntityRelationPaths(rightNode, paths, currentIdx + i + chunkOffset, graph, entities, relations, settings);
  });

  return chunkOffset + nodesToRight.length - 1; // offset
};

export function calculateQueryLogic(
  node: SerializedNode<LogicNodeAttributes>,
  graph: QueryMultiGraph,
  logics: SerializedNode<LogicNodeAttributes>[]
): AllLogicStatement {
  if (!node?.attributes) throw Error('Malformed Graph! Node has no attributes');
  let connectionsToLeft = graph.edges.filter((e) => e.target === node.key);

  let ret = [...node.attributes.logic.logic].map((l) => {
    if (typeof l !== 'string') throw Error('Malformed Graph! Logic node has no logic');
    if (!node.attributes) throw Error('Malformed Graph! Logic node has no attributes');

    if (l.includes('@')) {
      // @ means it needs to fetch data from connection
      const inputRefIdx = node.attributes.logic.inputs.findIndex((input, i) => input.name === l.replace('@', '')); // fetches the corresponding element from input definition
      const inputRef = node.attributes.logic.inputs[inputRefIdx];
      if (!inputRef) throw Error('Malformed Graph! Logic node has incorrect input reference');

      const connectionToInputRef = connectionsToLeft.find((c) => c?.attributes?.targetHandleData.attributeName === inputRef.name);
      if (!connectionToInputRef) {
        // Not connected, search for set or default value
        let val = node.attributes.inputs?.[inputRef.name] || inputRef.default;
        if (inputRef.type === 'string') {
          if (val) {
            val = `\"${val}\"`;
          } else {
            val = `\".*\"`; // Empty means allow anything
          }
        }
        // console.log('val', val, inputRef, node);
        return val;
      } else if (connectionToInputRef.attributes?.sourceHandleData.nodeType === QueryElementTypes.Logic) {
        // Is connected to another logic node
        const leftLogic = logics.find((r) => r.key === connectionToInputRef.attributes?.sourceHandleData.nodeId);
        if (!leftLogic) throw Error('Malformed Graph! Logic node is connected but has no logic data');
        return calculateQueryLogic(leftLogic, graph, logics);
      } else {
        if (!connectionToInputRef.attributes?.sourceHandleData)
          throw Error('Malformed Graph! Logic node is connected but has no sourceHandleData');
        // Is connected to entity or relation node
        return `@${connectionToInputRef.attributes.sourceHandleData.nodeId}.${connectionToInputRef.attributes.sourceHandleData.attributeName}`;
      }
    } else {
      return l;
    }
  });

  return ret as AllLogicStatement;
}

function queryLogicUnion(graphLogicChunks: AllLogicStatement[]): AllLogicStatement | undefined {
  if (graphLogicChunks.length === 1) return graphLogicChunks[0];
  else if (graphLogicChunks.length > 1) {
    return ['And', graphLogicChunks[0], queryLogicUnion(graphLogicChunks.slice(1)) || '0'];
  }
  return undefined;
}

/**
 * Converts the ReactFlow query to a json data structure to send the query to the backend.
 * @returns {BackendQueryFormat} A JSON object in the `JSONFormat`.
 */
export function Query2BackendQuery(
  saveStateID: string,
  graph: QueryMultiGraph,
  settings: QueryBuilderSettings,
  ml: ML = mlDefaultState
): BackendQueryFormat {
  let query: BackendQueryFormat = {
    saveStateID: saveStateID,
    query: [],
    machineLearning: [],
    limit: settings.limit,
    return: ['*'], // TODO
    cached: false,
  };

  Object.keys(ml).forEach((mlType) => {
    if (ml[mlType as MLTypes].enabled) {
      query.machineLearning.push({ type: mlType as MLTypes });
    }
  });

  let entities = graph.nodes.filter((n) => n?.attributes?.type === QueryElementTypes.Entity) as SerializedNode<EntityNodeAttributes>[];
  let relations = graph.nodes.filter((n) => n?.attributes?.type === QueryElementTypes.Relation) as SerializedNode<RelationNodeAttributes>[];

  const graphologyQuery = Graph.from(graph);
  graphologyQuery
    .filterNodes((n, att) => att?.type == QueryElementTypes.Logic)
    .forEach((n) => {
      graphologyQuery.dropNode(n);
    }); // Remove all logic nodes from the graph for cycle test
  if (hasCycle(graphologyQuery)) {
    const cycles = entities.map((entity, i) => {
      return allSimplePaths(graphologyQuery, entity.key, entity.key);
    });
    cycles.forEach((cycles_inner, i) => {
      cycles_inner.forEach((cycle, j) => {
        const origin = cycle[0];
        const target = cycle[cycle.length - 2];
        const newOrigin = graphologyQuery.addNode(origin + 'cycle', graphologyQuery.getNodeAttributes(origin));
        const edgeAttributes = graphologyQuery.getEdgeAttributes(target, origin);
        graphologyQuery.dropEdge(target, origin);
        graphologyQuery.addEdge(target, newOrigin, edgeAttributes);
      });
    });

    return Query2BackendQuery(saveStateID, graphologyQuery.export(), settings, ml);
  }
  // Chunk extraction: traverse graph to find all paths of logic between relations and entities
  let graphSequenceChunks: QueryGraphNodes[][] = [];
  let graphSequenceLogicChunks: QueryGraphNodes[][] = [];
  let graphSequenceChunksIdMap: Record<string, [number, number]> = {};
  let chunkOffset = 0;

  let entitiesEmptyLeftHandle = entities.filter((n) => !graph.edges.some((e) => e.target === n.key));
  let relationsEmptyLeftHandle = relations.filter((n) => !graph.edges.some((e) => e.target === n.key));
  // let entitiesEmptyRightHandle = entities.filter((n) => !n?.attributes?.rightRelationHandleId);
  entitiesEmptyLeftHandle.map((entity, i) => {
    // start with all entities that have no left handle, which means it "starts" a logic
    chunkOffset += traverseEntityRelationPaths(entity, graphSequenceChunks, i + chunkOffset, graph, entities, relations, settings);
  });
  if (entitiesEmptyLeftHandle.length > 0) chunkOffset += entitiesEmptyLeftHandle.length;
  relationsEmptyLeftHandle.map((relation, i) => {
    // then, for all relations that have no left handle, since they weren't accounted by the loop above
    chunkOffset += traverseEntityRelationPaths(relation, graphSequenceChunks, i + chunkOffset, graph, entities, relations, settings);
  });
  graphSequenceChunks.forEach((chunkSequence, i) => {
    chunkSequence.forEach((chunk, j) => {
      graphSequenceChunksIdMap[chunk.id || chunk.name || ''] = [i, j];
    });
  });

  // Logic pathways extraction: now traverse the graph again though the logic components to construct the logic chunks
  // The traversal is done in reverse order, starting from the logic pill's right handle connected to an entity or relation, and going backwards
  let logics = graph.nodes.filter((n) => n?.attributes?.type === QueryElementTypes.Logic) as SerializedNode<LogicNodeAttributes>[];
  let logicsRightHandleConnectedOutside = logics.filter((n) => {
    return graph.edges.some((e) => e.source === n.key && e.attributes?.targetHandleData.nodeType === QueryElementTypes.Entity);
  });
  let logicsRightHandleFinal = logics.filter((n) => {
    return !graph.edges.some((e) => e.source === n.key);
  });
  let graphLogicChunks = logicsRightHandleFinal.map((node) => calculateQueryLogic(node, graph, logics));
  query.logic = queryLogicUnion(graphLogicChunks);

  // Logic pathways extraction: now traverse the graph again though the logic components to construct the logic chunks

  // console.log('logics', logics);
  // console.log('entitiesEmptyLeftHandle', entitiesEmptyLeftHandle);
  // console.log('relationsEmptyLeftHandle', relationsEmptyLeftHandle);
  // console.log('graphSequenceChunks', graphSequenceChunks);
  // console.log('graphLogicChunks', graphLogicChunks);
  // console.log('logicsRightHandleConnectedOutside', logicsRightHandleConnectedOutside);
  // console.log('logicsRightHandleFinal', logicsRightHandleFinal);
  // console.log('graphSequenceChunksIdMap', graphSequenceChunksIdMap);
  if (!graphSequenceChunks || graphSequenceChunks.length === 0 || graphSequenceChunks?.[0].length === 0) return query;

  const processConnection = (chunk: QueryGraphNodes[], position: number): RelationStruct | NodeStruct => {
    const currNode = chunk[position];

    if (currNode.type === QueryElementTypes.Relation) {
      const _currNode = currNode as RelationNodeAttributes;
      const ret: RelationStruct = {
        ID: _currNode.id,
        label: _currNode.name || undefined,
        depth: _currNode.depth,
        direction: !_currNode.direction || _currNode.direction === 'right' ? 'TO' : _currNode.direction === 'left' ? 'FROM' : 'BOTH',
        node: chunk.length === position + 1 ? undefined : (processConnection(chunk, position + 1) as NodeStruct),
      };
      return ret;
    } else if (currNode.type === QueryElementTypes.Entity) {
      const ret: NodeStruct = {
        ID: currNode?.id,
        label: currNode?.name,
        // logic: LogicStruct[];
        // subQuery?: QueryStruct;
        // export: ExportNodeStruct[];
        relation: chunk.length === position + 1 ? undefined : (processConnection(chunk, position + 1) as RelationStruct),
      };
      return ret;
    }

    throw Error('Malformed Chunks! ' + chunk + position);
  };

  query.query = graphSequenceChunks.map((chunk, i) => {
    const ret: QueryStruct = {
      ID: 'path_' + i, //TODO: chunk[0].id ||
      node: processConnection(chunk, 0),
    };
    return ret;
  });

  console.debug('%cNew processed query', 'color: aquamarine', graph, query);

  return query;
}

// function processConnectionFromRelation(
//   c: SerializedEdge,
//   entities: EntityNodeAttributes[],
//   relations: RelationNodeAttributes[]
// ): NodeStruct {
//   if (c.attributes?.targetHandle === null) throw Error('Malformed Graph! One or more edges of a relation node do not exist');
//   const targetHandleData = fromHandleId(c.attributes?.targetHandle);

//   if (targetHandleData.nodeType === QueryElementTypes.Entity) {
//     const targetEntity = entities.find((r) => r.id === c.target);
//   } else if (targetHandleData.nodeType === QueryElementTypes.Relation) {
//     const targetRelation = relations.find((r) => r.id === c.target);
//   } else if (targetHandleData.nodeType === QueryElementTypes.Logic) {
//   }
// }
