/**
 * This program has been developed by students from the bachelor Computer Science at
 * Utrecht University within the Software Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */

import { getGroupName } from './ResultNodeLinkParserUseCase';
import { GraphQueryResult, Node } from '@graphpolaris/shared/lib/data-access';
import { MultiGraph } from 'graphology';

import {
  HyperEdgeI,
  HyperEdgeRange,
  PaohvisAxisInfo,
  PaohvisData,
  PaohvisFilters,
  PaohvisNodeInfo,
  PaohvisNodeOrder,
  Relation,
  ValueType,
  AugmentedEdgeAttributes,
  AugmentedNodeAttributesGraph,
  connectionFromTo,
} from '../types';
import AttributeFilterUsecase, { getIds } from './AttributesFilterUseCase';
import SortUseCase from './SortUseCase';
import { getWidthOfText, getUniqueArrays, findConnectionsNodes, buildGraphology } from './utils';
import style from '../paohvis.module.scss';

type Index = number;

/**
 * This parser is used to parse the incoming query result to the format that's needed to make the Paohvis table.
 */
export class ToPaohvisDataParserUseCase {
  private queryResult: GraphQueryResult;
  private xAxisNodeGroup: string;
  private yAxisNodeGroup: string;
  private paohvisFilters: PaohvisFilters;

  public constructor(queryResult: GraphQueryResult) {
    let limitedQueryResult: GraphQueryResult = {
      ...queryResult,
      nodes: queryResult.nodes.slice(0, 50),
      edges: queryResult.edges,
    };
    //this.queryResult = limitedQueryResult;
    this.queryResult = queryResult;

    this.xAxisNodeGroup = '';
    this.yAxisNodeGroup = '';
    this.paohvisFilters = { nodeFilters: [], edgeFilters: [] };
  }
  /**
   * Parses query results to the format that's needed to make a Paohvis table.
   * @param axisInfo is the information that's needed to parse everything to the correct axis.
   * @param nodeOrder is the order in which the nodes should be parsed.
   * @returns the information that's needed to make a Paohvis table.
   */
  public parseQueryResult(
    axisInfo: PaohvisAxisInfo,
    nodeOrder: PaohvisNodeOrder,
    dimensions: {
      current: any;
      //width: number;
      //heigth: number;
    },
    entityVerticalListed: string,
    entityHorizontal: string,
    entityVertical: string,
    //graphStructure: MultiGraph
  ): PaohvisData {
    this.setAxesNodeGroups(axisInfo, entityHorizontal, entityVertical);
    /*
    console.log('nodeOrder: ', nodeOrder);
    console.log('axisInfo ', axisInfo);
    console.log('entityHorizontal', entityHorizontal);
    console.log('entityVertical', entityVertical);
    */
    const augmentedNodes: AugmentedNodeAttributesGraph[] = this.queryResult.nodes.map((node: Node) => ({
      id: node.id,
      attributes: node.attributes,
      label: node.label,
    }));

    const augmentedEdges: AugmentedEdgeAttributes[] = this.queryResult.edges.map((edge: any) => ({
      id: edge.id,
      to: edge.to,
      from: edge.from,
      attributes: edge.attributes,
      label: edge.label,
    }));

    const graphStruc: MultiGraph = buildGraphology({ nodes: Object.values(augmentedNodes), edges: Object.values(augmentedEdges) });

    // potentially remove
    const filteredData = AttributeFilterUsecase.applyFilters(this.queryResult, this.paohvisFilters);

    // filter unnecessary node groups and relations
    const nodes = filteredData.nodes.filter((node) => {
      const nodeType = getGroupName(node);
      //return nodeType === this.xAxisNodeGroup || nodeType === this.yAxisNodeGroup;
      return nodeType === this.yAxisNodeGroup;
    });

    //const nodeListAttr: any[] = nodes.map((node) => node.attributes[entityVerticalListed]);
    //console.log(nodeListAttr);
    const nodesID: string[] = nodes.map((node) => node.id);
    const edges = filteredData.edges.filter((edge) => edge.label === axisInfo.relation.collection);
    const edgesConnection = findConnectionsNodes(nodesID, nodesID, graphStruc, this.yAxisNodeGroup, edges);

    // get hyperEdgeDegree
    const hyperEdgeDegree: Record<string, number> = ToPaohvisDataParserUseCase.GetHyperEdgeDegreeDict(
      nodes,
      edges,
      axisInfo.isYAxisEntityEqualToRelationFrom,
    );

    //parse nodes
    const rowInfo: PaohvisNodeInfo = ToPaohvisDataParserUseCase.parseNodes(
      nodes,
      hyperEdgeDegree,
      nodeOrder,
      this.yAxisNodeGroup,
      this.xAxisNodeGroup,
      entityVerticalListed,
    );

    //rowInfo.rowLabels[0] = 'mmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmm';
    const maxSizeAllowed = Math.round(0.17931 * dimensions.current.width);
    let nodeListAttr: any[] = [];

    if (entityVerticalListed == 'id') {
      nodeListAttr = nodes.map((node) => node.id);
    } else {
      nodeListAttr = nodes.map((node) => node.attributes[entityVerticalListed]);
    }

    // loops over the rowColumn text, gets the text width and computes the maximum length.
    //let maxLabelWidth = calcMaxRowLabelWidth(rowInfo.rowLabels, maxSizeAllowed);
    let maxLabelWidth = calcMaxRowLabelWidth(nodeListAttr, maxSizeAllowed);
    //parse hyperEdges
    //const filteredEdges = filterUnusedEdges(getIds(nodes), edges);
    const resultHyperEdgeRanges = ToPaohvisDataParserUseCase.parseHyperEdgeRanges(
      nodes,
      edgesConnection,
      axisInfo,
      rowInfo,
      entityVerticalListed,
    );
    //const resultHyperEdgeRanges = ToPaohvisDataParserUseCase.parseHyperEdgeRanges(nodes, filteredEdges, axisInfo, rowInfo);

    let rowLabelOutput = [];

    if (entityVerticalListed == 'id') {
      rowLabelOutput = rowInfo.rowLabels;
    } else {
      rowLabelOutput = nodeListAttr;
    }

    SortUseCase.sortHyperEdges(resultHyperEdgeRanges);

    return {
      rowLabelsList: rowLabelOutput,
      //rowLabels: rowInfo.rowLabels,
      rowLabels: rowLabelOutput,
      hyperEdgeRanges: resultHyperEdgeRanges,
      maxRowLabelWidth: maxLabelWidth,
    };
  }

  /**
   * Sets the x-axis and y-axis node groups from the given PaohvisAxisInfo in the parser.
   * @param axisInfo is the new PaohvisAxisInfo that will be used.
   */
  private setAxesNodeGroups(axisInfo: PaohvisAxisInfo, entityHorizontal: string, entityVertical: string): void {
    const relation: Relation = axisInfo.relation;

    //this.xAxisNodeGroup = relation.to;
    //this.yAxisNodeGroup = relation.from;

    this.xAxisNodeGroup = entityHorizontal;
    this.yAxisNodeGroup = entityVertical;
    /*
    if (axisInfo.isYAxisEntityEqualToRelationFrom) {
      this.xAxisNodeGroup = relation.to;
      this.yAxisNodeGroup = relation.from;
    } else {
      this.xAxisNodeGroup = relation.from;
      this.yAxisNodeGroup = relation.to;
    }
    */
  }

  /**
   * This parses the nodes to get the information that's needed to parse the hyperedges.
   * @param nodes are the nodes from the query result.
   * @param hyperEdgeDegree is the dictionary where you can find how many edges connected from the node.
   * @param nodeOrder is the order in which the nodes should be sorted.
   * @param yAxisNodeType is the type of nodes that should be on the y-axis.
   * @param xAxisNodeType is the type of nodes that should be on the x-axis.
   * @returns the information that's needed to parse the hyperedges.
   */
  private static parseNodes(
    nodes: Node[],
    hyperEdgeDegree: Record<string, number>,
    nodeOrder: PaohvisNodeOrder,
    yAxisNodeType: string,
    xAxisNodeType: string,
    entityVerticalListed: string,
  ): PaohvisNodeInfo {
    //const rowNodes = filterRowNodes(nodes, hyperEdgeDegree, yAxisNodeType);
    const rowNodes = nodes; // to list all nodes available, even if they do not have connections

    SortUseCase.sortNodes(nodeOrder, rowNodes, hyperEdgeDegree, entityVerticalListed);

    const rowLabels = getIds(rowNodes);

    //make dictionary for finding the index of a row
    const yNodesIndexDict: Record<string, number> = {};
    let yNodeIndexCounter = 0;
    for (let i = 0; i < rowNodes.length; i++) {
      yNodesIndexDict[rowNodes[i].id] = yNodeIndexCounter;
      yNodeIndexCounter++;
    }

    const xNodesAttributesDict: Record<string, any> = getXNodesAttributesDict(yAxisNodeType, xAxisNodeType, nodes);

    return {
      rowLabels: rowLabels,
      xNodesAttributesDict: xNodesAttributesDict,
      yNodesIndexDict: yNodesIndexDict,
    };
  }

  /**
   * Makes a dictionary where you can find how many edges are connected to the nodes.
   * @param nodes are the nodes where the edges should be counted for.
   * @param edges should be used to count how many edges are connected to the nodes.
   * @param isEntityRelationFrom is to decide if you need to count the from's or the to's of the edge.
   * @returns a dictionary where you can find how many edges are connected to the nodes.
   */
  private static GetHyperEdgeDegreeDict(nodes: Node[], edges: connectionFromTo[], isEntityRelationFrom: boolean): Record<string, number> {
    const hyperEdgeDegreeDict: Record<string, number> = {};

    //initialize dictionary
    nodes.forEach((node) => (hyperEdgeDegreeDict[node.id] = 0));

    //count node appearance frequencies
    edges.forEach((edge) => {
      if (isEntityRelationFrom) hyperEdgeDegreeDict[edge.from]++;
      else hyperEdgeDegreeDict[edge.to]++;
    });

    return hyperEdgeDegreeDict;
  }

  /**
   * Parses the edges to make hyperedge ranges for the Paohvis table.
   * @param nodes the unused nodes should already be filtered out.
   * @param edges the unused edges should already be filtered out.
   * @param axisInfo is the information that's needed to parse the edges to their respective hyperedge range.
   * @param rowInfo is the information about the nodes that's needed to parse the edges to their respective hyperedge range.
   * @returns the hyperedge ranges that will be used by the Paohvis table.
   */

  private static parseHyperEdgeRanges(
    nodes: Node[],
    //edges: connectionFromTo[],
    edges: connectionFromTo[],
    axisInfo: PaohvisAxisInfo,
    rowInfo: PaohvisNodeInfo,
    entityVerticalListed: string,
  ): HyperEdgeRange[] {
    if (nodes.length == 0 || edges.length == 0) return [];

    const resultHyperEdgeRanges: HyperEdgeRange[] = [];

    //is used to find on which index of HyperEdgeRanges the edge should be added
    const attributeRangesIndexDict: Record<string, Index> = {};
    let attributeRangeCounter = 0;

    //is used to find the index to which hyperedge it belongs to (in the array hyperEdgeRanges[i].hyperEdges)
    //dict[hyperEdgeRangeLabel] gives a dict where you can find with dict[edge.to] the index where it belongs to
    const hyperEdgesDict: Record<string, Record<string, Index>> = {};

    const xAxisAttributeType: string = axisInfo.selectedAttribute.name;
    //const xAxisAttributeType: string = 'No attribute was selected';

    //is used to find the node attributes for the x-axis
    const xNodesAttributesDict: Record<string, any> = rowInfo.xNodesAttributesDict;

    //is used to find the index of nodes in rowLabels
    const yNodesIndexDict: Record<string, number> = rowInfo.yNodesIndexDict;

    let edgeDirection: string;
    let edgeDirectionOpposite: string;

    /*
      Building paohvis data structure for rendering
      Check DEV331
      Option1: if two nodes have same connection, creates an fake hyperedge on paohvis - const isFromOnYAxis = false;
      Option2: there is no artificial hyperedge - const isFromOnYAxis = true;

    */

    //const isFromOnYAxis = axisInfo.isYAxisEntityEqualToRelationFrom; // boolean to decide whether: from (true) or to (false)
    const isFromOnYAxis = true;
    //const collection = axisInfo.relation.collection;
    /*
    console.log('nodes ', nodes);
    console.log('edges ', edges);
    console.log('axisInfo.selectedAttribute ', axisInfo.selectedAttribute);
    console.log(' rowInfo.xNodesAttributesDict ', rowInfo.xNodesAttributesDict);
    */
    /*
      Create resultHyperEdgeRanges.hyperEdges structure
        for each attribute creates a matrix:
      hyperedgesXnodes
      [ 0: [0,0,1,1]]
      [ 1: [0,1,0,1]]
      [ 2: [1,0,0,1]]
    */
    // loop through all edges - and creates hyperedge structure - does not fill it
    for (let i = 0; i < edges.length; i++) {
      const edge = edges[i];

      //console.log('iteration ', i, xAxisAttributeType, edge);

      ({ edgeDirection, edgeDirectionOpposite } = getEdgeDirections(isFromOnYAxis, edge));
      let edgeIndexInResult: number = yNodesIndexDict[edgeDirection];
      let edgeIndexInResultOpposite: number = yNodesIndexDict[edgeDirectionOpposite];

      // check if the chosen attribute is an attribute of the edge or the node
      let attribute: any; // attribute selected on the panel

      /*
      console.log(
        'inf: ',
        axisInfo.selectedAttribute.origin,
        AttributeOrigin.relation,
        xAxisAttributeType
        //edge.attributes[xAxisAttributeType]
      );

      console.log(
        'cond1 ',
        axisInfo.selectedAttribute.origin,
        AttributeOrigin.relation,
        axisInfo.selectedAttribute.origin == AttributeOrigin.relation
      );
      */
      //console.log('cond2 ', xNodesAttributesDict);

      attribute = edge.attributes[xAxisAttributeType];
      /*
      if (axisInfo.selectedAttribute.origin == AttributeOrigin.relation) attribute = edge.attributes[xAxisAttributeType];
      else attribute = xNodesAttributesDict[edgeDirectionOpposite][xAxisAttributeType];
        */
      //console.log('xAxisAttributeType ', xAxisAttributeType, attribute);
      /*
      // look this for allowing characteristics
      if (axisInfo.selectedAttribute.origin == AttributeOrigin.relation) {
        //console.log('1 ');
        attribute = edge.attributes[xAxisAttributeType];
      } else {
        //console.log('2 ', edgeDirectionOpposite, xAxisAttributeType);
        attribute = xNodesAttributesDict[edgeDirectionOpposite][xAxisAttributeType];
      }
      */
      // if no edge attribute was selected, then all edges will be placed in one big hyperEdgeRange
      if (attribute == undefined) attribute = ValueType.noAttribute;
      //console.log('attributeRangesIndexDict ', attributeRangesIndexDict);

      // for grouping on the attributes
      if (attribute in attributeRangesIndexDict) {
        const rangeIndex = attributeRangesIndexDict[attribute]; // index of attribute
        const targetHyperEdges = resultHyperEdgeRanges[rangeIndex].hyperEdges; // specific hyperedge
        const targetHyperEdgeIndex = hyperEdgesDict[attribute][edgeDirectionOpposite];

        //console.log('rangeIndex ', rangeIndex, targetHyperEdges, ' hyperEdgesDict ', hyperEdgesDict);
        //console.log('hyperEdgesDict ', hyperEdgesDict);

        if (targetHyperEdgeIndex != undefined) {
          // hyperedge group already exists so add edge to group
          targetHyperEdges[targetHyperEdgeIndex].indices.push(edgeIndexInResult);
        } else {
          // create new hyperedge group
          hyperEdgesDict[attribute][edgeDirectionOpposite] = targetHyperEdges.length;

          // 'add edge to new hyperedge group'
          targetHyperEdges.push({
            indices: [edgeIndexInResult, edgeIndexInResultOpposite],
            frequencies: newFrequencies(rowInfo.rowLabels),
            nameToShow: edgeDirectionOpposite,
          });
        }
      } else {
        // create new attribute range
        attributeRangesIndexDict[attribute] = attributeRangeCounter;
        attributeRangeCounter++;

        hyperEdgesDict[attribute] = {};
        hyperEdgesDict[attribute][edgeDirectionOpposite] = 0;

        let label: string;
        if (xAxisAttributeType != ValueType.noAttribute && xAxisAttributeType != '') label = attribute.toString();
        else label = 'No Attr';
        const hyperEdge: HyperEdgeI = {
          indices: [edgeIndexInResult, edgeIndexInResultOpposite],
          frequencies: newFrequencies(rowInfo.rowLabels),
          nameToShow: edgeDirectionOpposite,
        };
        const hyperEdgeRange: HyperEdgeRange = {
          rangeText: label,
          hyperEdges: [hyperEdge],
        };
        resultHyperEdgeRanges.push(hyperEdgeRange);
      }
    }

    // filter hyperedges.
    //console.log('resultHyperEdgeRanges ', resultHyperEdgeRanges);

    let resultHyperEdgeRanges2: any = [];

    for (let indexGroups = 0; indexGroups < resultHyperEdgeRanges.length; indexGroups++) {
      const arrayCheckFrequencies: number[][] = [];
      const arrayCheckIndices: number[][] = [];
      for (let indexHyperEdge = 0; indexHyperEdge < resultHyperEdgeRanges[indexGroups].hyperEdges.length; indexHyperEdge++) {
        arrayCheckFrequencies.push(resultHyperEdgeRanges[indexGroups].hyperEdges[indexHyperEdge].frequencies);
        arrayCheckIndices.push(resultHyperEdgeRanges[indexGroups].hyperEdges[indexHyperEdge].indices);
      }

      const { uniqueArrays, indicesUnique } = getUniqueArrays(arrayCheckFrequencies);

      const hyperEdges: HyperEdgeI[] = [];

      for (let indexUniqueFrequencies = 0; indexUniqueFrequencies < uniqueArrays.length; indexUniqueFrequencies++) {
        const hyperEdge: HyperEdgeI = {
          indices: arrayCheckIndices[indicesUnique[indexUniqueFrequencies]],
          frequencies: uniqueArrays[indexUniqueFrequencies],
          nameToShow: 'NDY',
        };
        hyperEdges.push(hyperEdge);
      }

      const hyperEdgeRange: HyperEdgeRange = {
        rangeText: resultHyperEdgeRanges[indexGroups].rangeText,
        hyperEdges: hyperEdges,
      };

      resultHyperEdgeRanges2.push(hyperEdgeRange);
    }
    //console.log('resultHyperEdgeRanges2 ', resultHyperEdgeRanges2);
    //console.log('axisInfo.selectedAttribute.type ', axisInfo.selectedAttribute.type);
    SortUseCase.sortHyperEdgeRanges(resultHyperEdgeRanges2, axisInfo.selectedAttribute.type, entityVerticalListed);

    return resultHyperEdgeRanges2;
  }

  /** Sets new PaohvisFilters. */
  public setPaohvisFilters(filters: PaohvisFilters): void {
    this.paohvisFilters = filters;
  }
}

/** Calculates the width that's needed for the rowlabels. */
function calcMaxRowLabelWidth2(rowLabels: string[]) {
  const margin = 10;
  let maxLabelWidth = 0;
  rowLabels.forEach((rowLabel) => {
    const textWidth: number = getWidthOfText(rowLabel + '   ', style.tableFontFamily, style.tableFontSize, style.tableFontWeight);
    //console.log(rowLabel, textWidth);
    if (textWidth > maxLabelWidth) maxLabelWidth = textWidth;
  });
  return maxLabelWidth + margin;
}

function calcMaxRowLabelWidth(rowLabels: string[], maxLenthAllowed: number) {
  //const maxLenthAllowed = 200;
  //const margin = maxLenthAllowed * 0.05;
  let isMaxLengthAllowedOver = false;
  let maxLabelWidth = 0;
  rowLabels.forEach((rowLabel) => {
    //const textWidth: number = getWidthOfText(rowLabel + '   ', style.tableFontFamily, style.tableFontSize, style.tableFontWeight);

    let c = document.createElement('canvas');
    let ctx = c.getContext('2d') as CanvasRenderingContext2D;
    //console.log('style.tableFontSize ', style.tableFontSize);
    let fontspec = style.tableFontWeight + ' ' + style.tableFontSize + ' ' + style.tableFontFamily;
    //console.log(fontspec);
    if (ctx.font !== fontspec) ctx.font = fontspec;

    // Measure the width of the text
    let textWidth = ctx.measureText(rowLabel).width;

    if (textWidth >= maxLenthAllowed) {
      isMaxLengthAllowedOver = true;
      maxLabelWidth = maxLenthAllowed;
      // Short text

      let truncatedText = '';
      let ellipsisWidth = ctx.measureText('...').width;
      for (let i = 0; i < rowLabel.length; i++) {
        let truncatedWidth = ctx.measureText(truncatedText + rowLabel[i]).width + ellipsisWidth;
        if (truncatedWidth > maxLenthAllowed) {
          // Text exceeded maximum width, break the loop
          break;
        }
        truncatedText += rowLabel[i];
      }
      //rowLabel = truncatedText + '...'; // Update txt to contain truncated text with three dots added
      rowLabels[rowLabels.indexOf(rowLabel)] = truncatedText + '...'; //add truncated element
    } else {
      if (!isMaxLengthAllowedOver) {
        if (textWidth > maxLabelWidth) maxLabelWidth = textWidth;
      }
    }

    //console.log(rowLabel, textWidth);
    //if (textWidth > maxLabelWidth) maxLabelWidth = textWidth;
  });
  //return margin + maxLabelWidth + margin;
  return maxLabelWidth * 1.2;
}

/** Gets a dictionary where you can find the attributes that belong to the nodes on teh x-axis.  */
function getXNodesAttributesDict(yAxisNodeType: string, xAxisNodeType: string, nodes: Node[]) {
  const resultXNodesAttributesDict: Record<string, any> = {};
  // it goes to case1:
  if (yAxisNodeType == xAxisNodeType) nodes.forEach((node) => (resultXNodesAttributesDict[node!.id] = node.attributes));
  //console.log('cas2 ', yAxisNodeType == xAxisNodeType);
  else
    nodes.forEach((node) => {
      if (getGroupName(node) == xAxisNodeType) resultXNodesAttributesDict[node.id] = node.attributes;
    });
  return resultXNodesAttributesDict;
}

/**
 * Gets which direction the edge will be read from. The direction can be read from:
 * 'from => to' or 'to => from'.
 */
function getEdgeDirections(isFromToRelation: boolean, edge: connectionFromTo) {
  let edgeDirection: string;
  let edgeDirectionOpposite: string;
  if (isFromToRelation) {
    edgeDirection = edge.from;
    edgeDirectionOpposite = edge.to;
  } else {
    edgeDirection = edge.to;
    edgeDirectionOpposite = edge.from;
  }
  return { edgeDirection, edgeDirectionOpposite };
}

/** This makes an array filled with zeroes, which is used to count the frequencies of edges. */
function newFrequencies(rowLabels: string[]): number[] {
  return Array(rowLabels.length).fill(0);
}

/** Filters out nodes that have no edges and nodes that are not the specified type. */
function filterRowNodes(nodes: Node[], hyperEdgeDegree: Record<string, number>, rowNodeType: string): Node[] {
  return nodes.filter((node) => {
    return hyperEdgeDegree[node.id] > 0 && getGroupName(node) == rowNodeType;
  });
}
